"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const assert = require("assert");
const bitcode_1 = require("bitcode");
const constants = require("../constants");
const utils_1 = require("../utils");
var irTypes = bitcode_1.builder.types;
exports.irTypes = irTypes;
var irValues = bitcode_1.builder.values;
exports.irValues = irValues;
var IRSignature = irTypes.Signature;
exports.IRSignature = IRSignature;
var IRDeclaration = irValues.constants.Declaration;
exports.IRDeclaration = IRDeclaration;
var IRFunc = irValues.constants.Func;
exports.IRFunc = IRFunc;
var IRBasicBlock = irValues.BasicBlock;
exports.IRBasicBlock = IRBasicBlock;
var IRPhi = irValues.instructions.Phi;
exports.IRPhi = IRPhi;
var IRType = irTypes.Type;
exports.IRType = IRType;
var IRValue = irValues.Value;
exports.IRValue = IRValue;
class Compilation {
    constructor(prefix, root, properties, spans, options) {
        this.prefix = prefix;
        this.root = root;
        this.properties = properties;
        this.options = options;
        this.bitcode = new bitcode_1.Module();
        this.cstringCache = new Map();
        this.globalId = new utils_1.Identifier('g_');
        this.resumptionTargets = new Set();
        this.debugMethod = undefined;
        this.ir = this.bitcode.createBuilder();
        this.invariantGroup = this.ir.metadata([
            this.ir.metadata('llparse.invariant'),
        ]);
        this.state = this.ir.struct('state');
        this.signature = {
            callback: {
                match: this.ir.signature(constants.INT, [
                    this.state.ptr(),
                    constants.TYPE_POS,
                    constants.TYPE_ENDPOS,
                ]),
                span: this.ir.signature(constants.INT, [
                    this.state.ptr(),
                    constants.TYPE_POS,
                    constants.TYPE_ENDPOS,
                ]),
                value: this.ir.signature(constants.INT, [
                    this.state.ptr(),
                    constants.TYPE_POS,
                    constants.TYPE_ENDPOS,
                    constants.TYPE_MATCH,
                ]),
            },
            node: this.ir.signature(constants.TYPE_OUTPUT, [
                this.state.ptr(),
                constants.TYPE_POS,
                constants.TYPE_ENDPOS,
                constants.TYPE_MATCH,
            ]),
        };
        // Put most used fields first
        this.state.addField(constants.TYPE_INDEX, constants.STATE_INDEX);
        spans.forEach((span) => {
            this.state.addField(constants.TYPE_SPAN_POS, constants.STATE_SPAN_POS + span.index);
            if (span.callbacks.length > 1) {
                this.state.addField(this.signature.callback.span.ptr(), constants.STATE_SPAN_CB + span.index);
            }
        });
        this.state.addField(constants.TYPE_ERROR, constants.STATE_ERROR);
        this.state.addField(constants.TYPE_REASON, constants.STATE_REASON);
        this.state.addField(constants.TYPE_ERROR_POS, constants.STATE_ERROR_POS);
        this.state.addField(constants.TYPE_DATA, constants.STATE_DATA);
        this.state.addField(this.signature.node.ptr(), constants.STATE_CURRENT);
        for (const property of properties) {
            let ty;
            switch (property.ty) {
                case 'i8':
                    ty = constants.I8;
                    break;
                case 'i16':
                    ty = constants.I16;
                    break;
                case 'i32':
                    ty = constants.I32;
                    break;
                case 'i64':
                    ty = constants.I64;
                    break;
                case 'ptr':
                    ty = constants.PTR;
                    break;
                default: throw new Error(`Unsupported user type: "${property.ty}"`);
            }
            this.state.addField(ty, property.name);
        }
        this.state.finalize();
    }
    buildBitcode(init) {
        return this.bitcode.build();
    }
    buildHeader() {
        let res = '';
        const PREFIX = this.prefix.toUpperCase().replace(/[^a-z]/gi, '_');
        const DEFINE = this.options.headerGuard === undefined ?
            `INCLUDE_${PREFIX}_H_` : this.options.headerGuard;
        res += `#ifndef ${DEFINE}\n`;
        res += `#define ${DEFINE}\n`;
        res += '#ifdef __cplusplus\n';
        res += 'extern "C" {\n';
        res += '#endif\n';
        res += '\n';
        res += '#include <stdint.h>\n';
        res += '\n';
        // Structure
        res += `typedef struct ${this.prefix}_s ${this.prefix}_t;\n`;
        res += `struct ${this.prefix}_s {\n`;
        for (const field of this.state.fields) {
            let ty;
            if (field.name === constants.STATE_REASON ||
                field.name === constants.STATE_ERROR_POS) {
                ty = 'const char*';
            }
            else if (field.ty.isEqual(constants.I8)) {
                ty = 'int8_t';
            }
            else if (field.ty.isEqual(constants.I16)) {
                ty = 'int16_t';
            }
            else if (field.ty.isEqual(constants.I32)) {
                ty = 'int32_t';
            }
            else if (field.ty.isEqual(constants.I64)) {
                ty = 'int64_t';
            }
            else if (field.name === constants.STATE_CURRENT ||
                field.ty.isEqual(constants.PTR) ||
                field.ty.isEqual(this.signature.callback.span)) {
                ty = 'void*';
            }
            else {
                throw new Error(`Unknown state property type: "${field.ty.typeString}"`);
            }
            res += `  ${ty} ${field.name};\n`;
        }
        res += '};\n';
        res += '\n';
        res += `int ${this.prefix}_init(${this.prefix}_t* s);\n`;
        res += `int ${this.prefix}_execute(${this.prefix}_t* s, ` +
            'const char* p, const char* endp);\n';
        res += '\n';
        res += '#ifdef __cplusplus\n';
        res += '}  /* extern "C" */\n';
        res += '#endif\n';
        res += `#endif  /* ${DEFINE} */\n`;
        return res;
    }
    // Arguments
    stateArg(bb) {
        return bb.parent.getArgument(constants.ARG_STATE);
    }
    posArg(bb) {
        return bb.parent.getArgument(constants.ARG_POS);
    }
    endPosArg(bb) {
        return bb.parent.getArgument(constants.ARG_ENDPOS);
    }
    matchArg(bb) {
        return bb.parent.getArgument(constants.ARG_MATCH);
    }
    // State fields
    indexField(bb) {
        return this.stateField(bb, constants.STATE_INDEX);
    }
    currentField(bb) {
        return this.stateField(bb, constants.STATE_CURRENT);
    }
    errorField(bb) {
        return this.stateField(bb, constants.STATE_ERROR);
    }
    reasonField(bb) {
        return this.stateField(bb, constants.STATE_REASON);
    }
    errorPosField(bb) {
        return this.stateField(bb, constants.STATE_ERROR_POS);
    }
    spanPosField(bb, index) {
        return this.stateField(bb, constants.STATE_SPAN_POS + index);
    }
    spanCbField(bb, index) {
        return this.stateField(bb, constants.STATE_SPAN_CB + index);
    }
    stateField(bb, name) {
        const state = this.stateArg(bb);
        const GEP_OFF = constants.GEP_OFF;
        const index = this.state.lookupField(name).index;
        return bb.getelementptr(state, GEP_OFF.val(0), GEP_OFF.val(index), true);
    }
    // Globals
    cstring(value) {
        if (this.cstringCache.has(value)) {
            return this.cstringCache.get(value);
        }
        const res = this.addGlobalConst('cstr', this.ir.cstring(value));
        this.cstringCache.set(value, res);
        return res;
    }
    blob(value) {
        return this.addGlobalConst('blob', this.ir.blob(value));
    }
    addGlobalConst(name, init) {
        const uniqueName = this.globalId.id(name).name;
        const glob = this.ir.global(init.ty.ptr(), uniqueName, init);
        glob.linkage = 'internal';
        glob.markConstant();
        this.bitcode.add(glob);
        return glob;
    }
    declareFunction(signature, name) {
        const decl = signature.declareFunction(name);
        this.bitcode.add(decl);
        return decl;
    }
    defineFunction(signature, name, paramNames) {
        const fn = signature.defineFunction(name, paramNames);
        this.bitcode.add(fn);
        return fn;
    }
    // Miscellaneous
    addResumptionTarget(decl) {
        this.resumptionTargets.add(decl);
    }
    getResumptionTargets() {
        return Array.from(this.resumptionTargets);
    }
    branch(bb, condition, weights) {
        const onTrue = bb.parent.createBlock('on_true');
        const onFalse = bb.parent.createBlock('on_false');
        const br = bb.branch(condition, onTrue, onFalse);
        if (weights !== undefined) {
            const WEIGHT = constants.BRANCH_WEIGHT;
            br.metadata.set('prof', this.ir.metadata([
                this.ir.metadata('branch_weights'),
                this.ir.metadata(WEIGHT.val(this.toProfWeight(weights.onTrue))),
                this.ir.metadata(WEIGHT.val(this.toProfWeight(weights.onFalse))),
            ]));
        }
        return { onTrue, onFalse };
    }
    switch(bb, value, keys, weights) {
        const otherwise = bb.parent.createBlock('otherwise');
        const cases = keys.map((key) => {
            return {
                block: bb.parent.createBlock(`case_${key}`),
                value: value.ty.toInt().val(key),
            };
        });
        const br = bb.switch(value, otherwise, cases);
        if (weights !== undefined) {
            const WEIGHT = constants.BRANCH_WEIGHT;
            assert.strictEqual(weights.cases.length, keys.length);
            const list = [
                this.ir.metadata('branch_weights'),
                this.ir.metadata(WEIGHT.val(this.toProfWeight(weights.otherwise))),
            ];
            for (const weight of weights.cases) {
                list.push(this.ir.metadata(WEIGHT.val(this.toProfWeight(weight))));
            }
            br.metadata.set('prof', this.ir.metadata(list));
        }
        return { otherwise, cases: cases.map((c) => c.block) };
    }
    truncate(bb, value, toTy, isSigned = false) {
        const fromTy = value.ty;
        assert(toTy.isInt());
        assert(fromTy.isInt());
        const fromITy = fromTy.toInt();
        const toITy = toTy.toInt();
        let res;
        // Same type!
        if (fromITy.isEqual(toITy)) {
            res = value;
            // Extend
        }
        else if (fromITy.width < toITy.width) {
            if (isSigned) {
                res = bb.cast('sext', value, toITy);
            }
            else {
                res = bb.cast('zext', value, toITy);
            }
            // Truncate
        }
        else {
            assert(fromITy.width > toITy.width);
            res = bb.cast('trunc', value, toITy);
        }
        return res;
    }
    debug(bb, message) {
        if (this.options.debug === undefined) {
            return bb;
        }
        // Lazily declare debug method
        if (this.debugMethod === undefined) {
            const sig = this.ir.signature(this.ir.void(), [
                this.state.ptr(),
                constants.TYPE_POS,
                constants.TYPE_ENDPOS,
                constants.TYPE_DEBUG_MSG,
            ]);
            this.debugMethod = this.declareFunction(sig, this.options.debug);
        }
        const str = this.cstring(message);
        const GEP_OFF = constants.GEP_OFF;
        const cast = bb.getelementptr(str, GEP_OFF.val(0), GEP_OFF.val(0), true);
        bb.call(this.debugMethod, [
            this.stateArg(bb),
            this.posArg(bb),
            this.endPosArg(bb),
            cast,
        ]);
        return bb;
    }
    // Internals
    toProfWeight(weight) {
        // Completely ad-hoc
        if (weight === 'likely') {
            return 0x10000;
        }
        else if (weight === 'unlikely') {
            return 1;
        }
        else {
            return weight;
        }
    }
}
exports.Compilation = Compilation;
//# sourceMappingURL=index.js.map