"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const assert = require("assert");
const debugAPI = require("debug");
const llparse_builder_1 = require("llparse-builder");
const compilerCode = require("../code");
const constants_1 = require("../constants");
const match_sequence_1 = require("../match-sequence");
const compiler = require("../node");
const span_1 = require("../span");
const compilerTransform = require("../transform");
const utils_1 = require("../utils");
const trie_1 = require("./trie");
const debug = debugAPI('llparse:translator');
class Translator {
    constructor(prefix, options, spans) {
        this.prefix = prefix;
        this.id = new utils_1.Identifier(this.prefix + '__n_');
        this.codeId = new utils_1.Identifier(this.prefix + '__c_');
        this.map = new Map();
        this.spanMap = new Map();
        this.codeCache = new Map();
        this.matchSequenceCache = new Map();
        this.options = {
            maxTableElemWidth: options.maxTableElemWidth === undefined ?
                constants_1.DEFAULT_TRANSLATOR_MAX_TABLE_WIDTH : options.maxTableElemWidth,
            minTableSize: options.minTableSize === undefined ?
                constants_1.DEFAULT_TRANSLATOR_MIN_TABLE_SIZE : options.minTableSize,
        };
        assert(0 < this.options.maxTableElemWidth, 'Invalid `maxTableElemWidth`, must be positive');
        this.spans = spans.concurrency.map((concurrent, index) => {
            const span = new span_1.Span(index, concurrent.map((apiSpan) => {
                return this.translateCode(apiSpan.callback);
            }));
            for (const apiSpan of concurrent) {
                this.spanMap.set(apiSpan, span);
            }
            return span;
        });
    }
    translate(node) {
        if (this.map.has(node)) {
            return this.map.get(node);
        }
        let result;
        const id = () => this.id.id(node.name);
        // Instantiate target class
        if (node instanceof llparse_builder_1.node.Error) {
            result = new compiler.Error(id(), node.code, node.reason);
        }
        else if (node instanceof llparse_builder_1.node.Pause) {
            result = new compiler.Pause(id(), node.code, node.reason);
        }
        else if (node instanceof llparse_builder_1.node.Consume) {
            result = new compiler.Consume(id(), node.field);
        }
        else if (node instanceof llparse_builder_1.node.SpanStart) {
            result = new compiler.SpanStart(id(), this.spanMap.get(node.span), this.translateCode(node.span.callback));
        }
        else if (node instanceof llparse_builder_1.node.SpanEnd) {
            result = new compiler.SpanEnd(id(), this.spanMap.get(node.span), this.translateCode(node.span.callback));
        }
        else if (node instanceof llparse_builder_1.node.Invoke) {
            result = new compiler.Invoke(id(), this.translateCode(node.code));
        }
        else if (node instanceof llparse_builder_1.node.Match) {
            result = this.translateMatch(node);
        }
        else {
            throw new Error(`Unknown node type for "${node.name}"`);
        }
        // Initialize result
        const otherwise = node.getOtherwiseEdge();
        if (Array.isArray(result)) {
            assert(node instanceof llparse_builder_1.node.Match);
            const match = node;
            // TODO(indutny): move this to llparse-builder?
            assert.notStrictEqual(otherwise, undefined, `Node "${node.name}" has no \`.otherwise()\``);
            // Assign otherwise to every node of Trie
            if (otherwise !== undefined) {
                for (const child of result) {
                    child.setOtherwise(this.translate(otherwise.node), otherwise.noAdvance);
                }
            }
            // Assign transform to every node of Trie
            const transform = this.translateTransform(match.getTransform());
            for (const child of result) {
                child.setTransform(transform);
            }
            assert(result.length >= 1);
            return result[0];
        }
        else if (result instanceof compiler.Node) {
            // Break loops
            this.map.set(node, result);
            if (otherwise !== undefined) {
                result.setOtherwise(this.translate(otherwise.node), otherwise.noAdvance);
            }
            else {
                // TODO(indutny): move this to llparse-builder?
                assert(node instanceof llparse_builder_1.node.Error, `Node "${node.name}" has no \`.otherwise()\``);
            }
            if (result instanceof compiler.Invoke) {
                for (const edge of node) {
                    result.addEdge(edge.key, this.translate(edge.node));
                }
            }
            else {
                assert.strictEqual(Array.from(node).length, 0);
            }
            return result;
        }
        else {
            throw new Error('Unreachable');
        }
    }
    translateMatch(node) {
        const trie = new trie_1.Trie(node.name);
        const otherwise = node.getOtherwiseEdge();
        const trieNode = trie.build(Array.from(node));
        if (trieNode === undefined) {
            return new compiler.Empty(this.id.id(node.name));
        }
        const children = [];
        this.translateTrie(node, trieNode, children);
        assert(children.length >= 1);
        return children;
    }
    translateTrie(node, trie, children) {
        if (trie instanceof trie_1.TrieEmpty) {
            assert(this.map.has(node));
            return this.translate(trie.node);
        }
        else if (trie instanceof trie_1.TrieSingle) {
            return this.translateSingle(node, trie, children);
        }
        else if (trie instanceof trie_1.TrieSequence) {
            return this.translateSequence(node, trie, children);
        }
        else {
            throw new Error('Unknown trie node');
        }
    }
    translateSingle(node, trie, children) {
        // See if we can apply TableLookup optimization
        const maybeTable = this.maybeTableLookup(node, trie, children);
        if (maybeTable !== undefined) {
            return maybeTable;
        }
        const single = new compiler.Single(this.id.id(node.name));
        children.push(single);
        // Break the loop
        if (!this.map.has(node)) {
            this.map.set(node, single);
        }
        for (const child of trie.children) {
            const childNode = this.translateTrie(node, child.node, children);
            single.addEdge({
                key: child.key,
                noAdvance: child.noAdvance,
                node: childNode,
                value: child.node instanceof trie_1.TrieEmpty ? child.node.value : undefined,
            });
        }
        return single;
    }
    maybeTableLookup(node, trie, children) {
        if (trie.children.length < this.options.minTableSize) {
            debug('not enough children of "%s" to allocate table, got %d need %d', node.name, trie.children.length, this.options.minTableSize);
            return undefined;
        }
        const targets = new Map();
        const bailout = !trie.children.every((child) => {
            if (!(child.node instanceof trie_1.TrieEmpty)) {
                debug('non-leaf trie child of "%s" prevents table allocation', node.name);
                return false;
            }
            const empty = child.node;
            // We can't pass values from the table yet
            if (empty.value !== undefined) {
                debug('value passing trie leaf of "%s" prevents table allocation', node.name);
                return false;
            }
            const target = empty.node;
            if (!targets.has(target)) {
                targets.set(target, {
                    keys: [child.key],
                    noAdvance: child.noAdvance,
                    trie: empty,
                });
                return true;
            }
            const existing = targets.get(target);
            // TODO(indutny): just use it as a sub-key?
            if (existing.noAdvance !== child.noAdvance) {
                debug('noAdvance mismatch in a trie leaf of "%s" prevents ' +
                    'table allocation', node.name);
                return false;
            }
            existing.keys.push(child.key);
            return true;
        });
        if (bailout) {
            return undefined;
        }
        // We've width limit for this optimization
        if (targets.size >= (1 << this.options.maxTableElemWidth)) {
            debug('too many different trie targets of "%s" for a table allocation', node.name);
            return undefined;
        }
        const table = new compiler.TableLookup(this.id.id(node.name));
        children.push(table);
        // Break the loop
        if (!this.map.has(node)) {
            this.map.set(node, table);
        }
        targets.forEach((target) => {
            const next = this.translateTrie(node, target.trie, children);
            table.addEdge({
                keys: target.keys,
                noAdvance: target.noAdvance,
                node: next,
            });
        });
        debug('optimized "%s" to a table lookup node', node.name);
        return table;
    }
    translateSequence(node, trie, children) {
        const matchSequence = this.createMatchSequence(node);
        matchSequence.addSequence(trie.select);
        const sequence = new compiler.Sequence(this.id.id(node.name), matchSequence, trie.select);
        children.push(sequence);
        // Break the loop
        if (!this.map.has(node)) {
            this.map.set(node, sequence);
        }
        const childNode = this.translateTrie(node, trie.child, children);
        const value = trie.child instanceof trie_1.TrieEmpty ?
            trie.child.value : undefined;
        sequence.setEdge(childNode, value);
        return sequence;
    }
    translateCode(code) {
        const prefixed = this.codeId.id(code.name).name;
        let res;
        if (code instanceof llparse_builder_1.code.IsEqual) {
            res = new compilerCode.IsEqual(prefixed, code.field, code.value);
        }
        else if (code instanceof llparse_builder_1.code.Load) {
            res = new compilerCode.Load(prefixed, code.field);
        }
        else if (code instanceof llparse_builder_1.code.MulAdd) {
            res = new compilerCode.MulAdd(prefixed, code.field, {
                base: code.options.base,
                max: code.options.max,
                signed: code.options.signed === undefined ? true : code.options.signed,
            });
        }
        else if (code instanceof llparse_builder_1.code.Or) {
            res = new compilerCode.Or(prefixed, code.field, code.value);
        }
        else if (code instanceof llparse_builder_1.code.Store) {
            res = new compilerCode.Store(prefixed, code.field);
        }
        else if (code instanceof llparse_builder_1.code.Test) {
            res = new compilerCode.Test(prefixed, code.field, code.value);
        }
        else if (code instanceof llparse_builder_1.code.Update) {
            res = new compilerCode.Update(prefixed, code.field, code.value);
            // External callbacks
        }
        else if (code instanceof llparse_builder_1.code.Match) {
            res = new compilerCode.Match(code.name);
        }
        else if (code instanceof llparse_builder_1.code.Span) {
            res = new compilerCode.Span(code.name);
        }
        else if (code instanceof llparse_builder_1.code.Value) {
            res = new compilerCode.Value(code.name);
        }
        else {
            throw new Error(`Unsupported code: "${code.name}"`);
        }
        // Re-use instances to build them just once
        if (this.codeCache.has(res.cacheKey)) {
            return this.codeCache.get(res.cacheKey);
        }
        this.codeCache.set(res.cacheKey, res);
        return res;
    }
    translateTransform(transform) {
        if (transform === undefined) {
            return new compilerTransform.ID();
        }
        else if (transform.name === 'to_lower_unsafe') {
            return new compilerTransform.ToLowerUnsafe();
        }
        else {
            throw new Error(`Unsupported transform: "${transform.name}"`);
        }
    }
    createMatchSequence(node) {
        const transform = this.translateTransform(node.getTransform());
        const cacheKey = transform === undefined ? '' : transform.name;
        if (this.matchSequenceCache.has(cacheKey)) {
            return this.matchSequenceCache.get(cacheKey);
        }
        const res = new match_sequence_1.MatchSequence(transform);
        this.matchSequenceCache.set(cacheKey, res);
        return res;
    }
}
exports.Translator = Translator;
//# sourceMappingURL=index.js.map