"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const assert = require("assert");
const buffer_1 = require("buffer");
const empty_1 = require("./empty");
exports.TrieEmpty = empty_1.TrieEmpty;
const node_1 = require("./node");
exports.TrieNode = node_1.TrieNode;
const sequence_1 = require("./sequence");
exports.TrieSequence = sequence_1.TrieSequence;
const single_1 = require("./single");
exports.TrieSingle = single_1.TrieSingle;
class Trie {
    constructor(name) {
        this.name = name;
    }
    build(edges) {
        if (edges.length === 0) {
            return undefined;
        }
        const internalEdges = [];
        for (const edge of edges) {
            internalEdges.push({
                key: edge.key,
                noAdvance: edge.noAdvance,
                node: edge.node,
                value: edge.value,
            });
        }
        return this.level(internalEdges, []);
    }
    level(edges, path) {
        const first = edges[0].key;
        const last = edges[edges.length - 1].key;
        let common = 0;
        const min = Math.min(first.length, last.length);
        // Leaf
        if (min === 0) {
            assert.strictEqual(edges.length, 1, `Duplicate entries in "${this.name}" at [ ${path.join(', ')} ]`);
            return new empty_1.TrieEmpty(edges[0].node, edges[0].value);
        }
        // Find the longest common sub-string
        for (; common < min; common++) {
            if (first[common] !== last[common]) {
                break;
            }
        }
        // Sequence
        if (common > 1) {
            return this.sequence(edges, first.slice(0, common), path);
        }
        // Single
        return this.single(edges, path);
    }
    slice(edges, off) {
        return edges.map((edge) => {
            return {
                key: edge.key.slice(off),
                noAdvance: edge.noAdvance,
                node: edge.node,
                value: edge.value,
            };
        }).sort((a, b) => {
            return a.key.compare(b.key);
        });
    }
    sequence(edges, prefix, path) {
        const sliced = this.slice(edges, prefix.length);
        const noAdvance = sliced.some((edge) => edge.noAdvance);
        assert(!noAdvance);
        const child = this.level(sliced, path.concat(prefix));
        return new sequence_1.TrieSequence(prefix, child);
    }
    single(edges, path) {
        const keys = new Map();
        for (const edge of edges) {
            const key = edge.key[0];
            if (keys.has(key)) {
                keys.get(key).push(edge);
            }
            else {
                keys.set(key, [edge]);
            }
        }
        const children = [];
        keys.forEach((subEdges, key) => {
            const sliced = this.slice(subEdges, 1);
            const subpath = path.concat(buffer_1.Buffer.from([key]));
            const noAdvance = subEdges.some((edge) => edge.noAdvance);
            const allSame = subEdges.every((edge) => edge.noAdvance === noAdvance);
            assert(allSame || subEdges.length === 0, 'Conflicting `.peek()` and `.match()` entries in ' +
                `"${this.name}" at [ ${subpath.join(', ')} ]`);
            children.push({
                key,
                noAdvance,
                node: this.level(sliced, subpath),
            });
        });
        return new single_1.TrieSingle(children);
    }
}
exports.Trie = Trie;
//# sourceMappingURL=index.js.map